@@ -57,7 +57,7 @@ use tokio::{
 };
 
 lazy_static! {
-    static ref VARS: Pool<Vec<(Chars, Value)>> = Pool::new(8, 2048);
+    static ref VARS: Pool<Vec<(Path, Chars, Value)>> = Pool::new(8, 2048);
     static ref REFS: Pool<Vec<(Path, Value)>> = Pool::new(8, 16384);
     static ref REFIDS: Pool<Vec<ExprId>> = Pool::new(8, 2048);
     static ref PKBUF: Pool<Vec<u8>> = Pool::new(8, 16384);
@@ -151,22 +151,36 @@ enum LcEvent {
 struct Lc {
     current_path: Path,
     db: db::Db,
-    var: FxHashMap<Chars, FxHashSet<ExprId>>,
-    sub: FxHashMap<SubId, FxHashSet<ExprId>>,
+    var: FxHashMap<Chars, FxHashMap<ExprId, usize>>,
+    sub: FxHashMap<SubId, FxHashMap<ExprId, usize>>,
     rpc: FxHashMap<Path, FxHashSet<ExprId>>,
-    refs: FxHashMap<Path, FxHashSet<ExprId>>,
+    refs: FxHashMap<Path, FxHashMap<ExprId, usize>>,
     rels: FxHashMap<Path, FxHashSet<ExprId>>,
     forward_refs: FxHashMap<ExprId, Refs>,
     subscriber: Subscriber,
     publisher: Publisher,
     sub_updates: mpsc::Sender<Pooled<Vec<(SubId, Event)>>>,
-    var_updates: Pooled<Vec<(Chars, Value)>>,
+    var_updates: Pooled<Vec<(Path, Chars, Value)>>,
     ref_updates: Pooled<Vec<(Path, Value)>>,
     by_id: FxHashMap<Id, Published>,
     by_path: HashMap<Path, Published>,
     events: mpsc::UnboundedSender<LcEvent>,
 }
 
+fn remove_eid_from_map<K: Hash + Eq>(
+    tbl: &mut FxHashMap<K, FxHashMap<ExprId, usize>>,
+    key: K,
+    expr_id: &ExprId,
+) {
+    if let Entry::Occupied(mut e) = tbl.entry(key) {
+        let set = e.get_mut();
+        set.remove(expr_id);
+        if set.is_empty() {
+            e.remove();
+        }
+    }
+}
+
 fn remove_eid_from_set<K: Hash + Eq>(
     tbl: &mut FxHashMap<K, FxHashSet<ExprId>>,
     key: K,
@@ -212,16 +226,16 @@ impl Lc {
     fn unref(&mut self, expr_id: ExprId) {
         if let Some(refs) = self.forward_refs.remove(&expr_id) {
             for path in refs.refs {
-                remove_eid_from_set(&mut self.refs, path, &expr_id);
+                remove_eid_from_map(&mut self.refs, path, &expr_id);
             }
             for path in refs.rpcs {
                 remove_eid_from_set(&mut self.rpc, path, &expr_id);
             }
             for id in refs.subs {
-                remove_eid_from_set(&mut self.sub, id, &expr_id);
+                remove_eid_from_map(&mut self.sub, id, &expr_id);
             }
             for name in refs.vars {
-                remove_eid_from_set(&mut self.var, name, &expr_id);
+                remove_eid_from_map(&mut self.var, name, &expr_id);
             }
         }
     }
@@ -253,30 +267,75 @@ impl Ctx for Lc {
     ) -> Dval {
         let dv = self.subscriber.durable_subscribe(path);
         dv.updates(flags, self.sub_updates.clone());
-        self.sub
+        *self
+            .sub
             .entry(dv.id())
-            .or_insert_with(|| HashSet::with_hasher(FxBuildHasher::default()))
-            .insert(ref_id);
+            .or_insert_with(|| HashMap::with_hasher(FxBuildHasher::default()))
+            .entry(ref_id)
+            .or_insert(0) += 1;
         self.forward_refs.entry(ref_id).or_insert_with(Refs::new).subs.insert(dv.id());
         dv
     }
 
-    fn ref_var(&mut self, name: Chars, ref_id: ExprId) {
-        self.var
+    fn unsubscribe(&mut self, _path: Path, dv: Dval, ref_id: ExprId) {
+        if let Entry::Occupied(mut etbl) = self.sub.entry(dv.id()) {
+            let tbl = etbl.get_mut();
+            if let Entry::Occupied(mut ecnt) = tbl.entry(ref_id) {
+                let cnt = ecnt.get_mut();
+                *cnt -= 1;
+                if *cnt == 0 {
+                    ecnt.remove();
+                    if tbl.is_empty() {
+                        etbl.remove();
+                    }
+                    if let Some(refs) = self.forward_refs.get_mut(&ref_id) {
+                        refs.subs.remove(&dv.id());
+                    }
+                }
+            }
+        }
+    }
+
+    // CR estokes: future optimization, use scope to avoid sending unecessary events
+    fn ref_var(&mut self, name: Chars, _scope: Path, ref_id: ExprId) {
+        *self
+            .var
             .entry(name.clone())
-            .or_insert_with(|| HashSet::with_hasher(FxBuildHasher::default()))
-            .insert(ref_id);
+            .or_insert_with(|| HashMap::with_hasher(FxBuildHasher::default()))
+            .entry(ref_id)
+            .or_insert(0) += 1;
         self.forward_refs.entry(ref_id).or_insert_with(Refs::new).vars.insert(name);
     }
 
+    fn unref_var(&mut self, name: Chars, _scope: Path, ref_id: ExprId) {
+        if let Entry::Occupied(mut etbl) = self.var.entry(name.clone()) {
+            let tbl = etbl.get_mut();
+            if let Entry::Occupied(mut ecnt) = tbl.entry(ref_id) {
+                let cnt = ecnt.get_mut();
+                *cnt -= 1;
+                if *cnt == 0 {
+                    ecnt.remove();
+                    if tbl.is_empty() {
+                        etbl.remove();
+                    }
+                    if let Some(refs) = self.forward_refs.get_mut(&ref_id) {
+                        refs.vars.remove(&name);
+                    }
+                }
+            }
+        }
+    }
+
     fn set_var(
         &mut self,
-        variables: &mut HashMap<Chars, Value>,
+        variables: &mut FxHashMap<Path, FxHashMap<Chars, Value>>,
+        local: bool,
+        scope: Path,
         name: Chars,
         value: Value,
     ) {
-        variables.insert(name.clone(), value.clone());
-        self.var_updates.push((name, value));
+        vm::store_var(variables, local, &scope, &name, value.clone());
+        self.var_updates.push((scope, name, value));
     }
 
     fn call_rpc(
@@ -365,32 +424,38 @@ impl Ref {
             | vm::Event::User(UserEv::Rel)
             | vm::Event::Netidx(_, _)
             | vm::Event::Rpc(_, _)
-            | vm::Event::Variable(_, _) => None,
+            | vm::Event::Variable(_, _, _) => None,
         }
     }
 
     fn set_ref(&mut self, ctx: &mut ExecCtx<Lc, UserEv>, path: Option<Chars>) {
         if let Some(path) = self.path.take() {
             let path = Path::from(path);
-            ctx.user
-                .refs
-                .entry(path.clone())
-                .or_insert_with(|| HashSet::with_hasher(FxBuildHasher::default()))
-                .remove(&self.id);
-            ctx.user
-                .forward_refs
-                .entry(self.id)
-                .or_insert_with(Refs::new)
-                .refs
-                .remove(&path);
+            if let Entry::Occupied(mut etbl) = ctx.user.refs.entry(path.clone()) {
+                let tbl = etbl.get_mut();
+                if let Entry::Occupied(mut ecnt) = tbl.entry(self.id) {
+                    let cnt = ecnt.get_mut();
+                    *cnt -= 1;
+                    if *cnt == 0 {
+                        ecnt.remove();
+                        if tbl.is_empty() {
+                            etbl.remove();
+                        }
+                        if let Some(refs) = ctx.user.forward_refs.get_mut(&self.id) {
+                            refs.refs.remove(&path);
+                        }
+                    }
+                }
+            }
         }
         if let Some(path) = path.clone() {
             let path = Path::from(path);
-            ctx.user
+            *ctx.user
                 .refs
                 .entry(path.clone())
-                .or_insert_with(|| HashSet::with_hasher(FxBuildHasher::default()))
-                .insert(self.id);
+                .or_insert_with(|| HashMap::with_hasher(FxBuildHasher::default()))
+                .entry(self.id)
+                .or_insert(0) += 1;
             ctx.user
                 .forward_refs
                 .entry(self.id)
@@ -404,7 +469,7 @@ impl Ref {
 
 impl Register<Lc, UserEv> for Ref {
     fn register(ctx: &mut ExecCtx<Lc, UserEv>) {
-        let f: InitFn<Lc, UserEv> = Arc::new(|ctx, from, id| {
+        let f: InitFn<Lc, UserEv> = Arc::new(|ctx, from, _, id| {
             let path = Ref::get_path(from);
             let current = Ref::get_current(ctx, &path);
             let mut t = Box::new(Ref { id, path: None, current });
@@ -500,7 +565,7 @@ impl Rel {
 
 impl Register<Lc, UserEv> for Rel {
     fn register(ctx: &mut ExecCtx<Lc, UserEv>) {
-        let f: InitFn<Lc, UserEv> = Arc::new(|ctx, from, id| {
+        let f: InitFn<Lc, UserEv> = Arc::new(|ctx, from, _, id| {
             let loc = ctx.user.current_path.clone();
             if let Some(table) = Path::dirname(&loc).and_then(Path::dirname) {
                 ctx.user
@@ -548,7 +613,7 @@ pub(crate) struct OnWriteEvent {
 
 impl Register<Lc, UserEv> for OnWriteEvent {
     fn register(ctx: &mut ExecCtx<Lc, UserEv>) {
-        let f: InitFn<Lc, UserEv> = Arc::new(|_, from, _| {
+        let f: InitFn<Lc, UserEv> = Arc::new(|_, from, _, _| {
             Box::new(OnWriteEvent { cur: None, invalid: from.len() > 0 })
         });
         ctx.functions.insert("event".into(), f);
@@ -572,7 +637,7 @@ impl Apply<Lc, UserEv> for OnWriteEvent {
     ) -> Option<Value> {
         self.invalid = from.len() > 0;
         match event {
-            vm::Event::Variable(_, _)
+            vm::Event::Variable(_, _, _)
             | vm::Event::Netidx(_, _)
             | vm::Event::Rpc(_, _)
             | vm::Event::User(UserEv::Ref(_, _))
@@ -778,8 +843,12 @@ impl Container {
         let on_write_expr_id =
             on_write_expr.as_ref().map(|e| e.id).unwrap_or_else(|_| ExprId::new());
         self.ctx.user.current_path = path.clone();
-        let formula_node = expr.map(|e| Node::compile(&mut self.ctx, e));
-        let on_write_node = on_write_expr.map(|e| Node::compile(&mut self.ctx, e));
+        let scope = Path::dirname(&path)
+            .map(|s| Path::from(String::from(s)))
+            .unwrap_or_else(Path::root);
+        let formula_node = expr.map(|e| Node::compile(&mut self.ctx, scope.clone(), e));
+        let on_write_node =
+            on_write_expr.map(|e| Node::compile(&mut self.ctx, scope.clone(), e));
         let value =
             formula_node.as_ref().ok().and_then(|n| n.current()).unwrap_or(Value::Null);
         let src_path = path.append(".formula");
@@ -915,7 +984,7 @@ impl Container {
             let r = REFS.take();
             for (path, value) in replace(&mut self.ctx.user.ref_updates, r).drain(..) {
                 if let Some(expr_ids) = self.ctx.user.refs.get(&path) {
-                    refs.extend(expr_ids.iter().copied());
+                    refs.extend(expr_ids.keys().copied());
                 }
                 self.update_expr_ids(
                     batch,
@@ -925,11 +994,17 @@ impl Container {
             }
             // update variable references
             let v = VARS.take();
-            for (name, value) in replace(&mut self.ctx.user.var_updates, v).drain(..) {
+            for (scope, name, value) in
+                replace(&mut self.ctx.user.var_updates, v).drain(..)
+            {
                 if let Some(expr_ids) = self.ctx.user.var.get(&name) {
-                    refs.extend(expr_ids.iter().copied());
+                    refs.extend(expr_ids.keys().copied());
                 }
-                self.update_expr_ids(batch, &mut refs, &vm::Event::Variable(name, value));
+                self.update_expr_ids(
+                    batch,
+                    &mut refs,
+                    &vm::Event::Variable(scope, name, value),
+                );
             }
             n += 1;
         }
@@ -962,7 +1037,7 @@ impl Container {
         for (id, event) in updates.drain(..) {
             if let Event::Update(value) = event {
                 if let Some(expr_ids) = self.ctx.user.sub.get(&id) {
-                    refs.extend(expr_ids.iter().copied());
+                    refs.extend(expr_ids.keys().copied());
                 }
                 self.update_expr_ids(batch, &mut refs, &vm::Event::Netidx(id, value));
             }
@@ -985,7 +1060,10 @@ impl Container {
             Ok(expr) => {
                 *expr_id = expr.id;
                 self.ctx.user.current_path = fifo.data_path.clone();
-                let node = Node::compile(&mut self.ctx, expr);
+                let scope = Path::dirname(&fifo.data_path)
+                    .map(|s| Path::from(String::from(s)))
+                    .unwrap_or_else(Path::root);
+                let node = Node::compile(&mut self.ctx, scope, expr);
                 let dv = node.current().unwrap_or(Value::Null);
                 let c = Compiled::Formula { node, data_id: fifo.data.id() };
                 self.compiled.insert(*expr_id, c);
@@ -1018,7 +1096,10 @@ impl Container {
             Ok(expr) => {
                 *expr_id = expr.id;
                 self.ctx.user.current_path = fifo.data_path.clone();
-                let node = Node::compile(&mut self.ctx, expr);
+                let scope = Path::dirname(&fifo.data_path)
+                    .map(|s| Path::from(String::from(s)))
+                    .unwrap_or_else(Path::root);
+                let node = Node::compile(&mut self.ctx, scope, expr);
                 self.compiled.insert(*expr_id, Compiled::OnWrite(node));
             }
             Err(_) => (), // CR estokes: log and report to user
